@@ -32,6 +32,7 @@
 #include "fcgid_bucket.h"
 #define FCGID_APPLY_TRY_COUNT 2
 #define FCGID_REQUEST_COUNT 32
+#define FCGID_BRIGADE_CLEAN_STEP 32
 
 static fcgid_procnode *apply_free_procnode(request_rec *r,
                                            fcgid_command * command)
@@ -315,7 +316,46 @@ handle_request_ipc(request_rec *r, int role,
     /* Check the script header first; return immediately on error. */
     if ((cond_status =
          ap_scan_script_header_err_core(r, sbuf, getsfunc_fcgid_BRIGADE,
-                                        brigade_stdout)) >= 400) {
+                                        brigade_stdout))) {
+        /*
+         * cond_status could be HTTP_NOT_MODIFIED in the case that the FCGI
+         * script does not set an explicit status and ap_meets_conditions,
+         * which is called by ap_scan_script_header_err_brigade, detects that
+         * the conditions of the requests are met and the response is
+         * not modified.
+         * In this case set r->status and return OK in order to prevent
+         * running through the error processing stack as this would
+         * break with mod_cache, if the conditions had been set by
+         * mod_cache itself to validate a stale entity.
+         * BTW: We circumvent the error processing stack anyway if the
+         * FCGI script set an explicit status code (whatever it is) and
+         * the only possible values for cond_status here are:
+         *
+         * HTTP_NOT_MODIFIED          (set by ap_meets_conditions)
+         * HTTP_PRECONDITION_FAILED   (set by ap_meets_conditions)
+         * HTTP_GATEWAY_TIME_OUT      (script timed out, returned no headers)
+         * HTTP_INTERNAL_SERVER_ERROR (if something went wrong during the
+         * processing of the response of the FCGI script, e.g broken headers
+         * or a crashed FCGI process).
+         */
+        if (cond_status == HTTP_NOT_MODIFIED) {
+            /* We need to remove our fcgid_filter before returning this
+             * status and code; otherwise, when ap_process_async_request()
+             * invokes ap_finalize_request_protocol() and that calls
+             * ap_pass_brigade(), fcgid_filter notices it has an empty
+             * brigade and returns without calling ap_pass_brigade() itself,
+             * which incorrectly circumvents the standard output filters.
+             */
+            ap_remove_output_filter(r->output_filters);
+
+            r->status = cond_status;
+            return OK;
+        }
+
+        return cond_status;
+    }
+
+    if (role == FCGI_AUTHORIZER) {
         return cond_status;
     }
 
@@ -323,32 +363,46 @@ handle_request_ipc(request_rec *r, int role,
     location = apr_table_get(r->headers_out, "Location");
 
     if (location && location[0] == '/' && r->status == 200) {
-        /* This redirect needs to be a GET no matter what the original 
-         * method was. 
+        /* This redirect needs to be a GET no matter what the original
+         * method was.
          */
         r->method = apr_pstrdup(r->pool, "GET");
         r->method_number = M_GET;
 
-        /* We already read the message body (if any), so don't allow 
-         * the redirected request to think it has one.  We can ignore 
-         * Transfer-Encoding, since we used REQUEST_CHUNKED_ERROR. 
+        /* We already read the message body (if any), so don't allow
+         * the redirected request to think it has one.  We can ignore
+         * Transfer-Encoding, since we used REQUEST_CHUNKED_ERROR.
          */
         apr_table_unset(r->headers_in, "Content-Length");
 
+        /* Setting this Location header value causes handle_request() to
+         * invoke ap_internal_redirect_handler(); that calls
+         * internal_internal_redirect() which sets the new sub-request's
+         * r->output_filters back to r->proto_output_filters before
+         * running the sub-request's handler.  Because we return here
+         * without invoking ap_pass_brigade(), our fcgid_filter is ignored.
+         */
         *location_ptr = location;
-        return HTTP_OK;
+        return OK;
     }
     else if (location && r->status == 200) {
-        /* XX Note that if a script wants to produce its own Redirect 
-         * body, it now has to explicitly *say* "Status: 302" 
+        /* XX Note that if a script wants to produce its own Redirect
+         * body, it now has to explicitly *say* "Status: 302"
+         */
+
+        /* This return code causes ap_process_async_request() to invoke
+         * ap_die(); that calls ap_send_error_response(), which resets
+         * r->output_filters back to r->proto_output_filters, thus removing
+         * our fcgid_filter from the output chain before making a final call
+         * to ap_finalize_request_protocol(), which passes the brigade to
+         * the standard output filters.
          */
         return HTTP_MOVED_TEMPORARILY;
     }
 
-    /* Now pass to output filter */
-    if (role == FCGI_RESPONDER
-        && (rv = ap_pass_brigade(r->output_filters,
-                                 brigade_stdout)) != APR_SUCCESS) {
+    /* Now pass any remaining response body data to output filters */
+    if ((rv = ap_pass_brigade(r->output_filters,
+                              brigade_stdout)) != APR_SUCCESS) {
         if (!APR_STATUS_IS_ECONNABORTED(rv)) {
             ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
                           "mod_fcgid: ap_pass_brigade failed in "
@@ -375,14 +429,12 @@ handle_request(request_rec * r, int role, fcgid_cmd_conf *cmd_conf,
     bucket_ctx->ipc.request = r;
     apr_pool_cleanup_register(r->pool, bucket_ctx,
                               bucket_ctx_cleanup, apr_pool_cleanup_null);
+    procmgr_init_spawn_cmd(&fcgi_request, r, cmd_conf);
 
     /* Try to get a connected ipc handle */
     for (i = 0; i < FCGID_REQUEST_COUNT; i++) {
         /* Apply a free process slot, send a spawn request if I can't get one */
         for (j = 0; j < FCGID_APPLY_TRY_COUNT; j++) {
-            /* Init spawn request */
-            procmgr_init_spawn_cmd(&fcgi_request, r, cmd_conf);
-
             bucket_ctx->ipc.connect_timeout =
                 fcgi_request.cmdopts.ipc_connect_timeout;
             bucket_ctx->ipc.communation_timeout =
@@ -405,7 +457,7 @@ handle_request(request_rec * r, int role, fcgid_cmd_conf *cmd_conf,
             }
 
             /* Send a spawn request if I can't get a process slot */
-            procmgr_post_spawn_cmd(&fcgi_request, r);
+            procmgr_send_spawn_cmd(&fcgi_request, r);
         }
 
         /* Connect to the fastcgi server */
@@ -465,7 +517,7 @@ handle_request(request_rec * r, int role, fcgid_cmd_conf *cmd_conf,
         ap_internal_redirect_handler(location, r);
     }
 
-    /* Retrun condition status */
+    /* Return condition status */
     return cond_status;
 }
 
@@ -479,20 +531,25 @@ static int add_request_body(request_rec *r, apr_pool_t *request_pool,
     FCGI_Header *stdin_request_header;
     fcgid_server_conf *sconf = ap_get_module_config(r->server->module_config,
                                                     &fcgid_module);
-    int seen_eos;
+    int seen_eos = 0;
 
     /* Stdin header and body */
-    /* I have to read all the request into memory before sending it 
-       to fastcgi application server, this prevents slow clients from 
-       keeping the server in processing too long. 
+    /* I have to read all the request into memory before sending it
+       to fastcgi application server, this prevents slow clients from
+       keeping the server in processing too long.
        But sometimes it's not acceptable (think about uploading a large attachment)
        Request will be stored in tmp file if the size larger than max_mem_request_len
      */
-    seen_eos = 0;
+
+    apr_bucket_brigade *input_brigade = apr_brigade_create(request_pool,
+                                                           r->connection->
+                                                           bucket_alloc);
+    apr_bucket_brigade *tmp_brigade = apr_brigade_create(request_pool,
+                                                           r->connection->
+                                                           bucket_alloc);
+
     do {
-        apr_bucket_brigade *input_brigade = apr_brigade_create(request_pool,
-                                                               r->connection->
-                                                               bucket_alloc);
+        int loop_counter = 0;
 
         if ((rv = ap_get_brigade(r->input_filters, input_brigade,
                                  AP_MODE_READBYTES,
@@ -501,17 +558,25 @@ static int add_request_body(request_rec *r, apr_pool_t *request_pool,
             ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
                           "mod_fcgid: can't get data from http client");
             apr_brigade_destroy(output_brigade);
+            apr_brigade_destroy(tmp_brigade);
             apr_brigade_destroy(input_brigade);
             return HTTP_INTERNAL_SERVER_ERROR;
         }
 
-        for (bucket_input = APR_BRIGADE_FIRST(input_brigade);
-             bucket_input != APR_BRIGADE_SENTINEL(input_brigade);
-             bucket_input = APR_BUCKET_NEXT(bucket_input)) {
+	
+
+        while ((bucket_input = APR_BRIGADE_FIRST(input_brigade)) != APR_BRIGADE_SENTINEL(input_brigade)) {
             const char *data;
             apr_size_t len;
             apr_bucket *bucket_stdin;
 
+            ++loop_counter;
+            if ((loop_counter % FCGID_BRIGADE_CLEAN_STEP) == 0) {
+                apr_brigade_cleanup(tmp_brigade);
+            }
+            APR_BUCKET_REMOVE(bucket_input);
+            APR_BRIGADE_INSERT_TAIL(tmp_brigade, bucket_input);
+
             if (APR_BUCKET_IS_EOS(bucket_input)) {
                 seen_eos = 1;
                 break;
@@ -525,6 +590,7 @@ static int add_request_body(request_rec *r, apr_pool_t *request_pool,
                 ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
                               "mod_fcgid: can't read request from HTTP client");
                 apr_brigade_destroy(input_brigade);
+                apr_brigade_destroy(tmp_brigade);
                 apr_brigade_destroy(output_brigade);
                 return HTTP_INTERNAL_SERVER_ERROR;
             }
@@ -629,6 +695,7 @@ static int add_request_body(request_rec *r, apr_pool_t *request_pool,
                 ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                               "mod_fcgid: header overflow");
                 apr_brigade_destroy(input_brigade);
+                apr_brigade_destroy(tmp_brigade);
                 apr_brigade_destroy(output_brigade);
                 return HTTP_INTERNAL_SERVER_ERROR;
             }
@@ -636,10 +703,14 @@ static int add_request_body(request_rec *r, apr_pool_t *request_pool,
             APR_BRIGADE_INSERT_TAIL(output_brigade, bucket_stdin);
         }
 
-        apr_brigade_destroy(input_brigade);
+        apr_brigade_cleanup(input_brigade);
+        apr_brigade_cleanup(tmp_brigade);
     }
     while (!seen_eos);
 
+    apr_brigade_destroy(input_brigade);
+    apr_brigade_destroy(tmp_brigade);
+
     /* Append an empty body stdin header */
     stdin_request_header = apr_bucket_alloc(sizeof(FCGI_Header),
                                             r->connection->bucket_alloc);
