@@ -131,9 +131,9 @@ void *merge_fcgid_server_config(apr_pool_t * p, void *basev, void *locv)
         merged->default_init_env = base->default_init_env;
     }
     else {
-        merged->default_init_env = 
+        merged->default_init_env =
           apr_table_copy(p, base->default_init_env);
-        apr_table_overlap(merged->default_init_env, 
+        apr_table_overlap(merged->default_init_env,
                           local->default_init_env,
                           APR_OVERLAP_TABLES_SET);
     }
@@ -146,8 +146,8 @@ void *merge_fcgid_server_config(apr_pool_t * p, void *basev, void *locv)
         merged->pass_headers = base->pass_headers;
     }
     else {
-        merged->pass_headers = 
-          apr_array_append(p, 
+        merged->pass_headers =
+          apr_array_append(p,
                            base->pass_headers,
                            local->pass_headers);
     }
@@ -569,7 +569,7 @@ static void add_envvar_to_table(apr_table_t *t, apr_pool_t *p,
 #endif
     apr_table_set(t, name, value ? value : "");
 }
-                                
+
 const char *add_default_env_vars(cmd_parms * cmd, void *dummy,
                                  const char *name, const char *value)
 {
@@ -621,9 +621,16 @@ const char *set_authenticator_info(cmd_parms * cmd, void *config,
     apr_status_t rv;
     apr_finfo_t finfo;
     fcgid_dir_conf *dirconfig = (fcgid_dir_conf *) config;
+    char **args;
 
-    /* Is the wrapper exist? */
-    if ((rv = apr_stat(&finfo, authenticator, APR_FINFO_NORM,
+    /* Get wrapper path */
+    apr_tokenize_to_argv(authenticator, &args, cmd->temp_pool);
+
+    if (*args == NULL || **args == '\0')
+        return "Invalid authenticator config";
+
+    /* Fetch only required file details inode + device */
+    if ((rv = apr_stat(&finfo, args[0], APR_FINFO_IDENT,
                        cmd->temp_pool)) != APR_SUCCESS) {
         return missing_file_msg(cmd->pool, "Authenticator", authenticator, rv);
     }
@@ -632,10 +639,8 @@ const char *set_authenticator_info(cmd_parms * cmd, void *config,
     dirconfig->authenticator_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->authenticator_info));
-    dirconfig->authenticator_info->cgipath =
-        apr_pstrdup(cmd->pool, authenticator);
-    dirconfig->authenticator_info->cmdline =
-        dirconfig->authenticator_info->cgipath;
+    dirconfig->authenticator_info->cgipath = apr_pstrdup(cmd->pool, args[0]);
+    dirconfig->authenticator_info->cmdline = authenticator;
     dirconfig->authenticator_info->inode = finfo.inode;
     dirconfig->authenticator_info->deviceid = finfo.device;
     return NULL;
@@ -670,9 +675,16 @@ const char *set_authorizer_info(cmd_parms * cmd, void *config,
     apr_status_t rv;
     apr_finfo_t finfo;
     fcgid_dir_conf *dirconfig = (fcgid_dir_conf *) config;
+    char **args;
 
-    /* Is the wrapper exist? */
-    if ((rv = apr_stat(&finfo, authorizer, APR_FINFO_NORM,
+    /* Get wrapper path */
+    apr_tokenize_to_argv(authorizer, &args, cmd->temp_pool);
+
+    if (*args == NULL || **args == '\0')
+        return "Invalid authorizer config";
+
+    /* Fetch only required file details inode + device */
+    if ((rv = apr_stat(&finfo, args[0], APR_FINFO_IDENT,
                        cmd->temp_pool)) != APR_SUCCESS) {
         return missing_file_msg(cmd->pool, "Authorizer", authorizer, rv);
     }
@@ -681,10 +693,8 @@ const char *set_authorizer_info(cmd_parms * cmd, void *config,
     dirconfig->authorizer_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->authorizer_info));
-    dirconfig->authorizer_info->cgipath =
-        apr_pstrdup(cmd->pool, authorizer);
-    dirconfig->authorizer_info->cmdline =
-        dirconfig->authorizer_info->cgipath;
+    dirconfig->authorizer_info->cgipath = apr_pstrdup(cmd->pool, args[0]);
+    dirconfig->authorizer_info->cmdline = authorizer;
     dirconfig->authorizer_info->inode = finfo.inode;
     dirconfig->authorizer_info->deviceid = finfo.device;
     return NULL;
@@ -719,9 +729,16 @@ const char *set_access_info(cmd_parms * cmd, void *config,
     apr_status_t rv;
     apr_finfo_t finfo;
     fcgid_dir_conf *dirconfig = (fcgid_dir_conf *) config;
+    char **args;
 
-    /* Is the wrapper exist? */
-    if ((rv = apr_stat(&finfo, access, APR_FINFO_NORM,
+    /* Get wrapper path */
+    apr_tokenize_to_argv(access, &args, cmd->temp_pool);
+
+    if (*args == NULL || **args == '\0')
+        return "Invalid access config";
+
+    /* Fetch only required file details inode + device */
+    if ((rv = apr_stat(&finfo, args[0], APR_FINFO_IDENT,
                        cmd->temp_pool)) != APR_SUCCESS) {
         return missing_file_msg(cmd->pool, "Access checker", access, rv);
     }
@@ -730,10 +747,8 @@ const char *set_access_info(cmd_parms * cmd, void *config,
     dirconfig->access_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->access_info));
-    dirconfig->access_info->cgipath =
-        apr_pstrdup(cmd->pool, access);
-    dirconfig->access_info->cmdline = 
-        dirconfig->access_info->cgipath;
+    dirconfig->access_info->cgipath = apr_pstrdup(cmd->pool, args[0]);
+    dirconfig->access_info->cmdline = access;
     dirconfig->access_info->inode = finfo.inode;
     dirconfig->access_info->deviceid = finfo.device;
     return NULL;
@@ -762,16 +777,81 @@ fcgid_cmd_conf *get_access_info(request_rec * r, int *authoritative)
     return NULL;
 }
 
+#ifdef WIN32
+/* FcgidWin32PreventOrphans
+ *
+ *   When Apache process gets recycled or shutdown abruptly, CGI processes 
+ *   spawned by mod_fcgid will get orphaned. Orphaning happens mostly when
+ *   Apache worker threads take more than 30 seconds to exit gracefully.
+ *
+ * Apache when run as windows service during shutdown/restart of service
+ * process (master/parent) will terminate child httpd process within 30
+ * seconds (refer \server\mpm\winnt\mpm_winnt.c:master_main() 
+ * int timeout = 30000; ~line#1142), therefore if Apache worker threads
+ * are too busy to react to Master's graceful exit signal within 30 seconds
+ * mod_fcgid cleanup routines will not get invoked (refer child_main()
+ * \server\mpm\winnt\child.c: apr_pool_destroy(pchild); ~line#2275)
+ * thereby orphaning all mod_fcgid spwaned CGI processes. Therefore we utilize
+ * Win32 JobObjects to clean up child processes automatically so that CGI
+ * processes are force-killed by win32 during abnormal mod_fcgid termination.
+ *
+ */
+const char *set_win32_prevent_process_orphans(cmd_parms *cmd, void *dummy,
+                                              int arg)
+{
+    server_rec *s = cmd->server;
+    fcgid_server_conf *config = ap_get_module_config(s->module_config,
+                                                     &fcgid_module);
+    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+#define SETUP_ERR_MSG "Error enabling CGI process orphan prevention"
+
+    if (err != NULL) {
+        return err;
+    }
+
+    if (arg && config->hJobObjectForAutoCleanup == NULL) {
+        /* Create Win32 job object to prevent CGI process oprhaning
+         */
+        JOBOBJECT_EXTENDED_LIMIT_INFORMATION job_info = { 0 };
+        config->hJobObjectForAutoCleanup = CreateJobObject(NULL, NULL);
+
+        if (config->hJobObjectForAutoCleanup == NULL) {
+            ap_log_perror(APLOG_MARK, APLOG_STARTUP|APLOG_ERR, apr_get_os_error(),
+                          cmd->pool, "mod_fcgid: unable to create job object.");
+            return SETUP_ERR_MSG;
+        }
+
+        /* Set job info so that all spawned CGI processes are associated
+         * with mod_fcgid
+         */
+        job_info.BasicLimitInformation.LimitFlags
+            = JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE;
+        if (SetInformationJobObject(config->hJobObjectForAutoCleanup,
+                                    JobObjectExtendedLimitInformation,
+                                    &job_info, sizeof(job_info)) == 0) {
+            ap_log_perror(APLOG_MARK, APLOG_STARTUP|APLOG_ERR, apr_get_os_error(),
+                          cmd->pool, "mod_fcgid: unable to set job object information.");
+            CloseHandle(config->hJobObjectForAutoCleanup);
+            config->hJobObjectForAutoCleanup = NULL;
+            return SETUP_ERR_MSG;
+        }
+    }
+
+    return NULL;
+}
+#endif /* WIN32*/
+
 const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
                                const char *wrapper_cmdline,
                                const char *extension,
                                const char *virtual)
 {
-    const char *path, *tmp;
+    const char *path;
     apr_status_t rv;
     apr_finfo_t finfo;
     fcgid_cmd_conf *wrapper = NULL;
     fcgid_dir_conf *config = (fcgid_dir_conf *) dirconfig;
+    char **args;
 
     /* Sanity checks */
 
@@ -790,13 +870,14 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
         return "Invalid wrapper file extension";
 
     /* Get wrapper path */
-    tmp = wrapper_cmdline;
-    path = ap_getword_white(cmd->temp_pool, &tmp);
+    apr_tokenize_to_argv(wrapper_cmdline, &args, cmd->temp_pool);
+    path = apr_pstrdup(cmd->pool, args[0]);
+
     if (path == NULL || *path == '\0')
         return "Invalid wrapper config";
 
-    /* Does the wrapper exist? */
-    if ((rv = apr_stat(&finfo, path, APR_FINFO_NORM,
+    /* Fetch only required file details inode + device */
+    if ((rv = apr_stat(&finfo, path, APR_FINFO_IDENT,
                        cmd->temp_pool)) != APR_SUCCESS) {
         return missing_file_msg(cmd->pool, "Wrapper", path, rv);
     }
@@ -912,13 +993,14 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
     cmdopts->min_class_process_count = DEFAULT_MIN_CLASS_PROCESS_COUNT;
     cmdopts->proc_lifetime = DEFAULT_PROC_LIFETIME;
     /* via pcalloc: cmdopts->initenv_key[0][0] = '\0'; */
-    
+
     cmdname = ap_getword_conf(cmd->pool, &args);
     if (!strlen(cmdname)) {
         return "A command must be specified for FcgidCmdOptions";
     }
 
-    rv = apr_stat(&finfo, cmdname, APR_FINFO_NORM, cmd->temp_pool);
+    /* Test only for file existence */
+    rv = apr_stat(&finfo, cmdname, APR_FINFO_MIN, cmd->temp_pool);
     if (rv != APR_SUCCESS) {
         return missing_file_msg(cmd->pool, "Command", cmdname, rv);
     }
@@ -928,23 +1010,23 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
     }
 
     while (*args) {
-        const char *option = ap_getword_white(cmd->pool, &args);
+        const char *option = ap_getword_conf(cmd->pool, &args);
         const char *val;
 
         /* TODO: Consider supporting BusyTimeout.
          */
 
         if (!strcasecmp(option, "ConnectTimeout")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "ConnectTimeout must have an argument";
             }
             cmdopts->ipc_connect_timeout = atoi(val);
             continue;
         }
-            
+
         if (!strcasecmp(option, "IdleTimeout")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "IdleTimeout must have an argument";
             }
@@ -956,7 +1038,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
             char *name;
             char *eql;
 
-            name = ap_getword_white(cmd->pool, &args);
+            name = ap_getword_conf(cmd->pool, &args);
             if (!strlen(name)) {
                 return "InitialEnv must have an argument";
             }
@@ -975,16 +1057,16 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
         }
 
         if (!strcasecmp(option, "IOTimeout")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "IOTimeout must have an argument";
             }
             cmdopts->ipc_comm_timeout = atoi(val);
             continue;
         }
-            
+
         if (!strcasecmp(option, "MaxProcesses")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "MaxProcesses must have an argument";
             }
@@ -993,7 +1075,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
         }
 
         if (!strcasecmp(option, "MaxProcessLifetime")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "MaxProcessLifetime must have an argument";
             }
@@ -1002,7 +1084,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
         }
 
         if (!strcasecmp(option, "MaxRequestsPerProcess")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "MaxRequestsPerProcess must have an argument";
             }
@@ -1011,7 +1093,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
         }
 
         if (!strcasecmp(option, "MinProcesses")) {
-            val = ap_getword_white(cmd->pool, &args);
+            val = ap_getword_conf(cmd->pool, &args);
             if (!strlen(val)) {
                 return "MinProcesses must have an argument";
             }
