@@ -18,14 +18,12 @@
 #ifndef FCGID_CONF_H
 #define FCGID_CONF_H
 
-#include "apr_general.h" /* stringify */
-
 #define MODFCGID_COPYRIGHT \
-  "Copyright 2010 The Apache Software Foundation."
+  "Copyright 2013 The Apache Software Foundation."
 
 #define MODFCGID_VERSION_MAJOR  2
 #define MODFCGID_VERSION_MINOR  3
-#define MODFCGID_VERSION_SUBVER 6
+#define MODFCGID_VERSION_SUBVER 9
 #define MODFCGID_VERSION_DEV    0
 
 #if MODFCGID_VERSION_DEV
@@ -34,6 +32,14 @@
 #define MODFCGID_VERSION_DEVSTR ""
 #endif
 
+/* APR_STRINGIFY is defined here, and also in apr_general.h, so wrap it */
+#ifndef APR_STRINGIFY
+/** Properly quote a value as a string in the C preprocessor */
+#define APR_STRINGIFY(n) APR_STRINGIFY_HELPER(n)
+/** Helper macro for APR_STRINGIFY */
+#define APR_STRINGIFY_HELPER(n) #n
+#endif
+
 #define MODFCGID_REVISION      APR_STRINGIFY(MODFCGID_VERSION_MAJOR) \
                            "." APR_STRINGIFY(MODFCGID_VERSION_MINOR) \
                            "." APR_STRINGIFY(MODFCGID_VERSION_SUBVER)
@@ -71,6 +77,10 @@ typedef struct {
     int termination_score;
     int time_score;
     int zombie_scan_interval;
+#ifdef WIN32
+    /* FcgidWin32PreventOrphans - Win32 CGI processes automatic cleanup */
+    HANDLE hJobObjectForAutoCleanup;
+#endif
     /* global or vhost
      * scalar values have corresponding _set field to aid merging
      */
@@ -249,12 +259,18 @@ const char *set_php_fix_pathinfo_enable(cmd_parms * cmd, void *dummy,
 const char *set_max_requests_per_process(cmd_parms * cmd, void *dummy,
                                          const char *arg);
 
+#ifdef WIN32
+const char *set_win32_prevent_process_orphans(cmd_parms *cmd, void *dummy,
+                                              int arg);
+#endif
+
 const char *set_cmd_options(cmd_parms *cmd, void *dummy,
                             const char *arg);
 
 void get_cmd_options(request_rec *r, const char *cmdpath,
                      fcgid_cmd_options *cmdopts, fcgid_cmd_env *cmdenv);
 
+
 AP_MODULE_DECLARE_DATA extern module fcgid_module;
 
 #endif
