@@ -19,6 +19,7 @@
 #define CORE_PRIVATE
 #include "httpd.h"
 #include "http_config.h"
+#include "apr_strings.h"
 
 #include "fcgid_pm.h"
 #include "fcgid_pm_main.h"
@@ -43,8 +44,8 @@ link_node_to_list(server_rec * main_server,
 static apr_time_t lastidlescan = 0;
 static void scan_idlelist(server_rec * main_server)
 {
-    /* 
-       Scan the idle list 
+    /*
+       Scan the idle list
        1. move all processes idle timeout to error list
        2. move all processes lifetime expired to error list
      */
@@ -137,7 +138,7 @@ static void scan_busylist(server_rec * main_server)
         last_active_time = current_node->last_active_time;
         if (apr_time_sec(now) - apr_time_sec(last_active_time) >
             (current_node->cmdopts.busy_timeout)) {
-            /* Protocol: 
+            /* Protocol:
                1. diewhy init with FCGID_DIE_KILLSELF
                2. Process manager set diewhy to FCGID_DIE_BUSY_TIMEOUT and gracefully kill process while busy timeout
                3. Process manager forced kill process while busy timeout and diewhy is FCGID_DIE_BUSY_TIMEOUT
@@ -157,8 +158,8 @@ static void scan_busylist(server_rec * main_server)
 static apr_time_t lastzombiescan = 0;
 static void scan_idlelist_zombie(server_rec * main_server)
 {
-    /* 
-       Scan the idle list 
+    /*
+       Scan the idle list
        1. pick up the node for scan(now-last_activ>g_zombie_scan_interval)
        2. check if it's zombie process
        3. if it's zombie process, wait() and return to free list
@@ -184,8 +185,8 @@ static void scan_idlelist_zombie(server_rec * main_server)
         return;
     lastzombiescan = now;
 
-    /* 
-       Check the list 
+    /*
+       Check the list
      */
     proc_table = proctable_get_table_array();
     previous_node = proctable_get_idle_list();
@@ -214,7 +215,7 @@ static void scan_idlelist_zombie(server_rec * main_server)
     }
     proctable_pm_unlock(main_server);
 
-    /* 
+    /*
        Now check every node in check list
        1) If it's zombie process, wait() and return to free list
        2) If it's not zombie process, link it to the tail of idle list
@@ -244,8 +245,8 @@ static void scan_idlelist_zombie(server_rec * main_server)
         current_node = next_node;
     }
 
-    /* 
-       Now link the check list back to the tail of idle list 
+    /*
+       Now link the check list back to the tail of idle list
      */
     if (check_list_header->next_index) {
         proctable_pm_lock(main_server);
@@ -267,7 +268,7 @@ static void scan_idlelist_zombie(server_rec * main_server)
 static apr_time_t lasterrorscan = 0;
 static void scan_errorlist(server_rec * main_server)
 {
-    /* 
+    /*
        kill() and wait() every node in error list
        put them back to free list after that
      */
@@ -281,6 +282,7 @@ static void scan_errorlist(server_rec * main_server)
     fcgid_server_conf *sconf =
         ap_get_module_config(main_server->module_config,
                              &fcgid_module);
+    int graceful_terminations = 0;
 
     /* Should I check the busy list? */
     if (procmgr_must_exit()
@@ -325,6 +327,7 @@ static void scan_errorlist(server_rec * main_server)
                               current_node->proc_pool);
         if (!dummy) {
             proc_kill_gracefully(current_node, main_server);
+            ++graceful_terminations;
             apr_pool_userdata_set("set", HAS_GRACEFUL_KILL,
                                   apr_pool_cleanup_null,
                                   current_node->proc_pool);
@@ -349,51 +352,113 @@ static void scan_errorlist(server_rec * main_server)
     }
     previous_node->next_index = temp_error_header.next_index;
     proctable_pm_unlock(main_server);
+
+    if (graceful_terminations) {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, main_server,
+                     "mod_fcgid: gracefully terminated %d processes",
+                     graceful_terminations);
+    }
 }
 
-static void kill_all_subprocess(server_rec * main_server)
+typedef enum action_t {DO_NOTHING, KILL_GRACEFULLY, KILL_FORCEFULLY,
+                       HARD_WAIT} action_t;
+
+static int reclaim_one_pid(server_rec *main_server, fcgid_procnode *proc,
+                           action_t action)
 {
-    size_t i;
     int exitcode;
     apr_exit_why_e exitwhy;
+    apr_wait_how_e wait_how = action == HARD_WAIT ? APR_WAIT : APR_NOWAIT;
+
+    if (apr_proc_wait(&proc->proc_id, &exitcode, &exitwhy,
+                      wait_how) != APR_CHILD_NOTDONE) {
+        proc->diewhy = FCGID_DIE_SHUTDOWN;
+        proc_print_exit_info(proc, exitcode, exitwhy,
+                             main_server);
+        proc->proc_pool = NULL;
+        return 1;
+    }
+
+    switch(action) {
+    case DO_NOTHING:
+    case HARD_WAIT:
+        break;
+
+    case KILL_GRACEFULLY:
+        proc_kill_gracefully(proc, main_server);
+        break;
+
+    case KILL_FORCEFULLY:
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, main_server,
+                     "FastCGI process %" APR_PID_T_FMT
+                     " still did not exit, "
+                     "terminating forcefully",
+                     proc->proc_id.pid);
+        proc_kill_force(proc, main_server);
+        break;
+    }
+
+    return 0;
+}
+
+static void kill_all_subprocess(server_rec *main_server)
+{
+    apr_time_t waittime = 1024 * 16;
+    size_t i, table_size = proctable_get_table_size();
+    int not_dead_yet;
+    int cur_action, next_action;
+    apr_time_t starttime = apr_time_now();
+    struct {
+        action_t action;
+        apr_time_t action_time;
+    } action_table[] = {
+        {DO_NOTHING,      0}, /* dummy entry for iterations where
+                              * we reap children but take no action
+                              * against stragglers
+                              */
+        {KILL_GRACEFULLY, apr_time_from_sec(0)},
+        {KILL_GRACEFULLY, apr_time_from_sec(1)},
+        {KILL_FORCEFULLY, apr_time_from_sec(8)},
+        {HARD_WAIT,       apr_time_from_sec(8)}
+    };
     fcgid_procnode *proc_table = proctable_get_table_array();
 
-    /* Kill gracefully */
-    for (i = 0; i < proctable_get_table_size(); i++) {
-        if (proc_table[i].proc_pool)
-            proc_kill_gracefully(&proc_table[i], main_server);
-    }
-    apr_sleep(apr_time_from_sec(1));
-
-    /* Kill with SIGKILL if it doesn't work */
-    for (i = 0; i < proctable_get_table_size(); i++) {
-        if (proc_table[i].proc_pool) {
-            if (apr_proc_wait(&(proc_table[i].proc_id), &exitcode, &exitwhy,
-                              APR_NOWAIT) != APR_CHILD_NOTDONE) {
-                proc_table[i].diewhy = FCGID_DIE_SHUTDOWN;
-                proc_print_exit_info(&proc_table[i], exitcode, exitwhy,
-                                     main_server);
-                apr_pool_destroy(proc_table[i].proc_pool);
-                proc_table[i].proc_pool = NULL;
-            }
-            else
-                proc_kill_force(&proc_table[i], main_server);
+    next_action = 1;
+    do {
+        apr_sleep(waittime);
+        /* don't let waittime get longer than 1 second; otherwise, we don't
+         * react quickly to the last child exiting, and taking action can
+         * be delayed
+         */
+        waittime = waittime * 4;
+        if (waittime > apr_time_from_sec(1)) {
+            waittime = apr_time_from_sec(1);
         }
-    }
 
-    /* Wait again */
-    for (i = 0; i < proctable_get_table_size(); i++) {
-        if (proc_table[i].proc_pool) {
-            if (apr_proc_wait(&(proc_table[i].proc_id), &exitcode, &exitwhy,
-                              APR_WAIT) != APR_CHILD_NOTDONE) {
-                proc_table[i].diewhy = FCGID_DIE_SHUTDOWN;
-                proc_print_exit_info(&proc_table[i], exitcode, exitwhy,
-                                     main_server);
-                apr_pool_destroy(proc_table[i].proc_pool);
-                proc_table[i].proc_pool = NULL;
+        /* see what action to take, if any */
+        if (action_table[next_action].action_time <= apr_time_now() - starttime) {
+            cur_action = next_action;
+            ++next_action;
+        }
+        else {
+            cur_action = 0; /* index of DO_NOTHING entry */
+        }
+
+        /* now see who is done */
+        not_dead_yet = 0;
+        for (i = 0; i < table_size; i++) {
+            if (!proc_table[i].proc_pool) {
+                continue; /* unused */
+            }
+
+            if (!reclaim_one_pid(main_server, &proc_table[i],
+                                 action_table[cur_action].action)) {
+                ++not_dead_yet;
             }
         }
-    }
+
+    } while (not_dead_yet &&
+             action_table[cur_action].action != HARD_WAIT);
 }
 
 /* This should be proposed as a stand-alone improvement to the httpd module,
@@ -492,8 +557,12 @@ fastcgi_spawn(fcgid_command * command, server_rec * main_server,
     /* Prepare to spawn */
     procnode->deviceid = command->deviceid;
     procnode->inode = command->inode;
-    procnode->share_grp_id = command->share_grp_id;
-    procnode->virtualhost = command->virtualhost;
+
+    /* no truncation should ever occur */
+    AP_DEBUG_ASSERT(sizeof procnode->cmdline > strlen(command->cmdline));
+    apr_cpystrn(procnode->cmdline, command->cmdline, sizeof procnode->cmdline);
+
+    procnode->vhost_id = command->vhost_id;
     procnode->uid = command->uid;
     procnode->gid = command->gid;
     procnode->start_time = procnode->last_active_time = apr_time_now();
@@ -537,7 +606,7 @@ fastcgi_spawn(fcgid_command * command, server_rec * main_server,
     /* Spawn the process now */
     /* XXX Spawn uses wrapper_cmdline, but log uses cgipath ? */
     if ((rv =
-         proc_spawn_process(command->wrapper_cmdline, &procinfo,
+         proc_spawn_process(command->cmdline, &procinfo,
                             procnode)) != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_WARNING, rv, main_server,
                      "mod_fcgid: spawn process %s error", command->cgipath);
@@ -553,7 +622,9 @@ fastcgi_spawn(fcgid_command * command, server_rec * main_server,
                           procnode, proctable_array);
         ap_log_error(APLOG_MARK, APLOG_INFO, 0, main_server,
                      "mod_fcgid: server %s:%s(%" APR_PID_T_FMT ") started",
-                     command->virtualhost, command->cgipath,
+                     command->server_hostname[0] ?
+                         command->server_hostname : "(unknown)",
+                     command->cgipath,
                      procnode->proc_id.pid);
         register_spawn(main_server, procnode);
     }
@@ -568,7 +639,7 @@ apr_status_t pm_main(server_rec * main_server, apr_pool_t * configpool)
             break;
 
         /* Wait for command */
-        if (procmgr_peek_cmd(&command, main_server) == APR_SUCCESS) {
+        if (procmgr_fetch_cmd(&command, main_server) == APR_SUCCESS) {
             if (is_spawn_allowed(main_server, &command))
                 fastcgi_spawn(&command, main_server, configpool);
 
