@@ -204,26 +204,14 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
     char **proc_environ;
     struct sockaddr_un unix_addr;
     apr_procattr_t *procattr = NULL;
-    int argc, len;
-    const char *wargv[APACHE_ARG_MAX + 1];
-    const char *word; /* For wrapper */
-    const char *tmp;
+    int len;
+    const char **wargv;
 
     /* Build wrapper args */
-    argc = 0;
-    tmp = cmdline;
-    while (1) {
-        word = ap_getword_white(procnode->proc_pool, &tmp);
-        if (word == NULL || *word == '\0')
-            break;
-        if (argc >= APACHE_ARG_MAX)
-            break;
-        wargv[argc++] = word;
-    }
-    wargv[argc] = NULL;
+    apr_tokenize_to_argv(cmdline, (char ***)&wargv, procnode->proc_pool);
 
-    /* 
-       Create UNIX domain socket before spawn 
+    /*
+       Create UNIX domain socket before spawn
      */
 
     /* Generate a UNIX domain socket file path */
@@ -249,6 +237,7 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
                 sizeof(procnode->socket_path));
 
     /* truncation already checked for in handler or FcgidWrapper parser */
+    AP_DEBUG_ASSERT(wargv[0] != NULL);
     AP_DEBUG_ASSERT(strlen(wargv[0]) < sizeof(procnode->executable_path));
     apr_cpystrn(procnode->executable_path, wargv[0],
                 sizeof(procnode->executable_path));
@@ -378,7 +367,7 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
 
     /* fork and exec now */
     /* Note, don't pass &(procnode->proc_id) to fcgid_create_privileged_process(),
-     * for it's a share memory address, both parent and child process may modify 
+     * for it's a share memory address, both parent and child process may modify
      * procnode->proc_id->pid, so sometimes it's 0 and sometimes it's >0
      */
     rv = fcgid_create_privileged_process(&tmpproc, wargv[0], wargv,
@@ -402,6 +391,7 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
     procnode->proc_id = tmpproc;
 
     if (rv != APR_SUCCESS) {
+        memset(&procnode->proc_id, 0, sizeof(procnode->proc_id));
         ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
                      "mod_fcgid: can't run %s", wargv[0]);
     }
@@ -414,6 +404,11 @@ static apr_status_t proc_kill_internal(fcgid_procnode *procnode, int sig)
     /* su as root before sending signal, for suEXEC */
     apr_status_t rv;
 
+    if (procnode->proc_id.pid == 0) {
+        /* procnode->proc_id.pid be 0 while fcgid_create_privileged_process() fail */
+        return APR_SUCCESS;
+    }
+
     if (ap_unixd_config.suexec_enabled && seteuid(0) != 0) {
 
         /* can't gain privileges to send signal (should not occur); do NOT
@@ -461,6 +456,7 @@ apr_status_t proc_wait_process(server_rec *main_server, fcgid_procnode *procnode
         /* Destroy pool */
         apr_pool_destroy(procnode->proc_pool);
         procnode->proc_pool = NULL;
+        memset(&procnode->proc_id, 0, sizeof(procnode->proc_id));
 
         return APR_CHILD_DONE;
     }
@@ -514,7 +510,6 @@ apr_status_t proc_connect_ipc(fcgid_procnode *procnode, fcgid_ipc *ipc_handle)
     fcgid_namedpipe_handle *handle_info;
     struct sockaddr_un unix_addr;
     apr_status_t rv;
-    apr_int32_t on = 1;
 
     /* Alloc memory for unix domain socket */
     ipc_handle->ipc_handle_info
@@ -548,10 +543,6 @@ apr_status_t proc_connect_ipc(fcgid_procnode *procnode, fcgid_ipc *ipc_handle)
         return APR_ECONNREFUSED;
     }
 
-    /* Set no delay option */
-    setsockopt(handle_info->handle_socket, IPPROTO_TCP, TCP_NODELAY,
-               (char *) &on, sizeof(on));
-
     /* Set nonblock option */
     if ((rv = set_socket_nonblock(handle_info->handle_socket)) != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_DEBUG, rv, ipc_handle->request,
@@ -775,9 +766,13 @@ apr_status_t proc_write_ipc(fcgid_ipc *ipc_handle,
     for (e = APR_BRIGADE_FIRST(output_brigade);
          e != APR_BRIGADE_SENTINEL(output_brigade);
          e = APR_BUCKET_NEXT(e)) {
-        /* Read bucket */
         apr_size_t len;
         const char* base;
+
+        if (APR_BUCKET_IS_METADATA(e)) {
+            continue;
+        }
+
         if ((rv = apr_bucket_read(e, &base, &len,
                                   APR_BLOCK_READ)) != APR_SUCCESS) {
             ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, ipc_handle->request,
