@@ -62,29 +62,17 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
                                 fcgid_procnode *procnode)
 {
     HANDLE *finish_event, listen_handle;
-    int bufused = 0;
     SECURITY_ATTRIBUTES SecurityAttributes;
+    fcgid_server_conf *sconf;
     apr_procattr_t *proc_attr;
     apr_status_t rv;
     apr_file_t *file;
-    char **proc_environ;
+    const char * const *proc_environ;
     char sock_path[FCGID_PATH_MAX];
-    int argc;
-    char *wargv[APACHE_ARG_MAX + 1], *word; /* For wrapper */
-    const char *tmp;
+    const char **wargv;
 
     /* Build wrapper args */
-    argc = 0;
-    tmp = cmdline;
-    while (1) {
-        word = ap_getword_white(procnode->proc_pool, &tmp);
-        if (word == NULL || *word == '\0')
-            break;
-        if (argc >= APACHE_ARG_MAX)
-            break;
-        wargv[argc++] = word;
-    }
-    wargv[argc] = NULL;
+    apr_tokenize_to_argv(cmdline, (char ***)&wargv, procnode->proc_pool);
 
     memset(&SecurityAttributes, 0, sizeof(SecurityAttributes));
 
@@ -112,7 +100,7 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
 
     /* Prepare the listen namedpipe file name (no check for truncation) */
     apr_snprintf(sock_path, sizeof sock_path,
-                 "\\\\.\\pipe\\fcgidpipe-%u.%lu",
+                 "\\\\.\\pipe\\fcgidpipe-%lu.%d",
                  GetCurrentProcessId(), g_process_counter++);
 
     /* Prepare the listen namedpipe handle */
@@ -135,7 +123,8 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
                 sizeof(procnode->executable_path));
 
     /* Build environment variables */
-    proc_environ = ap_create_environment(procnode->proc_pool,
+    proc_environ = (const char * const *)
+                   ap_create_environment(procnode->proc_pool,
                                          procinfo->proc_environ);
     if (!proc_environ) {
         ap_log_error(APLOG_MARK, APLOG_WARNING, apr_get_os_error(),
@@ -153,7 +142,7 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
         || (rv = apr_procattr_cmdtype_set(proc_attr, APR_PROGRAM))
                != APR_SUCCESS
         || (rv = apr_procattr_detach_set(proc_attr, 1)) != APR_SUCCESS
-        || (rv = apr_procattr_io_set(proc_attr, APR_NO_PIPE, 
+        || (rv = apr_procattr_io_set(proc_attr, APR_NO_PIPE,
                                      APR_NO_FILE, APR_NO_FILE)) != APR_SUCCESS
         || (rv = apr_os_file_put(&file, &listen_handle, 0,
                                  procnode->proc_pool)) != APR_SUCCESS
@@ -168,7 +157,8 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
 
     /* fork and exec now */
     rv = apr_proc_create(&(procnode->proc_id), wargv[0], wargv,
-                         proc_environ, proc_attr, procnode->proc_pool);
+                         proc_environ,
+                         proc_attr, procnode->proc_pool);
 
     /* OK, I created the process, now put it back to idle list */
     CloseHandle(listen_handle);
@@ -176,9 +166,25 @@ apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
     if (rv != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
                      "mod_fcgid: can't run %s", wargv[0]);
+        return rv;
     }
 
-    return rv;
+    /* FcgidWin32PreventOrphans feature */
+    sconf = ap_get_module_config(procinfo->main_server->module_config,
+                                 &fcgid_module);
+
+    if (sconf->hJobObjectForAutoCleanup != NULL) {
+        /* Associate cgi process to current process */
+        if (AssignProcessToJobObject(sconf->hJobObjectForAutoCleanup,
+                                     procnode->proc_id.hproc) == 0) {
+            ap_log_error(APLOG_MARK, APLOG_WARNING, apr_get_os_error(),
+                         procinfo->main_server,
+                         "mod_fcgid: unable to assign child process to "
+                         "job object");
+        }
+    }
+
+    return APR_SUCCESS;
 }
 
 apr_status_t proc_kill_gracefully(fcgid_procnode *procnode,
@@ -297,7 +303,7 @@ apr_status_t proc_connect_ipc(fcgid_procnode *procnode, fcgid_ipc *ipc_handle)
                           " server %" APR_PID_T_FMT " has been terminated",
                           procnode->proc_id.pid);
         else
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, apr_get_os_error(), 
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, apr_get_os_error(),
                           ipc_handle->request,
                           "mod_fcgid: can't connect to named pipe, FastCGI"
                           " server pid %" APR_PID_T_FMT,
@@ -351,7 +357,7 @@ apr_status_t proc_read_ipc(fcgid_ipc * ipc_handle, const char *buffer,
                                      &transferred, FALSE /* don't wait */ )
                 || transferred == 0) {
                 rv = apr_get_os_error();
-                ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, 
+                ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv,
                               ipc_handle->request,
                               "mod_fcgid: get overlap result error");
                 return rv;
@@ -381,14 +387,11 @@ apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
          bucket_request != APR_BRIGADE_SENTINEL(birgade_send);
          bucket_request = APR_BUCKET_NEXT(bucket_request))
     {
-        char *write_buf;
+        const char *write_buf;
         apr_size_t write_buf_len;
         apr_size_t has_write;
 
-        if (APR_BUCKET_IS_EOS(bucket_request))
-            break;
-
-        if (APR_BUCKET_IS_FLUSH(bucket_request))
+        if (APR_BUCKET_IS_METADATA(bucket_request))
             continue;
 
         if ((rv = apr_bucket_read(bucket_request, &write_buf, &write_buf_len,
@@ -415,7 +418,7 @@ apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
                               "mod_fcgid: can't write to pipe");
                 return rv;
             } else {
-                /* 
+                /*
                    it's ERROR_IO_PENDING on write
                  */
                 DWORD dwWaitResult =
