@@ -49,6 +49,7 @@ register_life_death(server_rec * main_server,
     struct fcgid_stat_node *previous_node, *current_node;
     fcgid_server_conf *sconf = ap_get_module_config(main_server->module_config,
                                                     &fcgid_module);
+    apr_time_t now = apr_time_now();
 
     if (!g_stat_pool || !procnode)
         abort();
@@ -67,31 +68,13 @@ register_life_death(server_rec * main_server,
         previous_node = current_node;
     }
 
-    if (current_node) {
-        /* Found the node */
-        apr_time_t now = apr_time_now();
-
-        /* Increase the score first */
-        if (life_or_death == REGISTER_LIFE) {
-            current_node->score += sconf->spawn_score;
-            current_node->process_counter++;
-        } else {
-            current_node->score += sconf->termination_score;
-            current_node->process_counter--;
+    if (!current_node) {         /* I can't find it, create one */
+        if (life_or_death == REGISTER_DEATH) {
+            ap_log_error(APLOG_MARK, APLOG_CRIT, 0, main_server,
+                         "stat node not found for exiting process %s",
+                         procnode->cmdline);
+            return;
         }
-
-        /* Decrease the score base on the time passing */
-        current_node->score -= 
-          sconf->time_score * 
-          (int)(apr_time_sec(now) - apr_time_sec(current_node->last_stat_time));
-
-        /* Make score reasonable */
-        if (current_node->score < 0)
-            current_node->score = 0;
-
-        current_node->last_stat_time = now;
-    } else {
-        /* I can't find it, create one */
         current_node = apr_pcalloc(g_stat_pool, sizeof(*current_node));
         current_node->deviceid = procnode->deviceid;
         current_node->inode = procnode->inode;
@@ -99,9 +82,9 @@ register_life_death(server_rec * main_server,
         current_node->vhost_id = procnode->vhost_id;
         current_node->uid = procnode->uid;
         current_node->gid = procnode->gid;
-        current_node->last_stat_time = apr_time_now();
+        current_node->last_stat_time = now;
         current_node->score = 0;
-        current_node->process_counter = 1;
+        current_node->process_counter = 0;
         current_node->max_class_process_count =
             procnode->cmdopts.max_class_process_count;
         current_node->min_class_process_count =
@@ -114,6 +97,26 @@ register_life_death(server_rec * main_server,
         else
             previous_node->next = current_node;
     }
+
+    /* Increase the score first */
+    if (life_or_death == REGISTER_LIFE) {
+        current_node->score += sconf->spawn_score;
+        current_node->process_counter++;
+    } else {
+        current_node->score += sconf->termination_score;
+        current_node->process_counter--;
+    }
+
+    /* Decrease the score based on elapsed time */
+    current_node->score -=
+        sconf->time_score *
+        (int)(apr_time_sec(now) - apr_time_sec(current_node->last_stat_time));
+
+    /* Make score reasonable */
+    if (current_node->score < 0)
+        current_node->score = 0;
+
+    current_node->last_stat_time = now;
 }
 
 void spawn_control_init(server_rec * main_server, apr_pool_t * configpool)
@@ -141,15 +144,17 @@ void register_spawn(server_rec * main_server, fcgid_procnode * procnode)
     g_total_process++;
 }
 
-/* 
-    Spawn control is base on such strategy:
-    1. Increate score if application is terminated
-    2. Increate score if application is spawned
-    3. Decrease score each second while score is positive
-    4. Negative spawn request if score is higher than up limit
-    5. Negative spawn request if total process count higher than up limit
-    6. Negative spawn request if process count of this class higher than up limit
-*/
+/*
+ * Spawn control strategy:
+ * 1. Add FcgidSpawnScore to score if application is terminated
+ * 2. Add FcgidTerminationScore to score if application is spawned
+ * 3. Subtract FcgidTimeScore from score each second while score is positive
+ * 4. Ignore spawn request if score is higher than FcgidSpawnScoreUpLimit
+ * 5. Ignore spawn request if total process count higher than
+ *    FcgidSpawnScoreUpLimit
+ * 6. Ignore spawn request if process count of this class is higher than
+ *    FcgidMaxProcessesPerClass
+ */
 int is_spawn_allowed(server_rec * main_server, fcgid_command * command)
 {
     struct fcgid_stat_node *current_node;
@@ -161,7 +166,7 @@ int is_spawn_allowed(server_rec * main_server, fcgid_command * command)
 
     /* Total process count higher than up limit? */
     if (g_total_process >= sconf->max_process_count) {
-        ap_log_error(APLOG_MARK, APLOG_INFO, 0, main_server,
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, main_server,
                      "mod_fcgid: %s total process count %d >= %d, skip the spawn request",
                      command->cgipath, g_total_process, sconf->max_process_count);
         return 0;
@@ -173,14 +178,14 @@ int is_spawn_allowed(server_rec * main_server, fcgid_command * command)
         if (current_node->inode == command->inode
             && current_node->deviceid == command->deviceid
             && !strcmp(current_node->cmdline, command->cmdline)
-            && current_node->vhost_id == sconf->vhost_id
+            && current_node->vhost_id == command->vhost_id
             && current_node->uid == command->uid
             && current_node->gid == command->gid)
             break;
     }
 
     if (!current_node) {
-        /* There are no existing processes for this class, so obviously 
+        /* There are no existing processes for this class, so obviously
          * no class-specific limits have been exceeded.
          */
         return 1;
@@ -188,8 +193,8 @@ int is_spawn_allowed(server_rec * main_server, fcgid_command * command)
     else {
         apr_time_t now = apr_time_now();
 
-        current_node->score -= 
-          sconf->time_score * 
+        current_node->score -=
+          sconf->time_score *
           (int)(apr_time_sec(now) - apr_time_sec(current_node->last_stat_time));
         current_node->last_stat_time = now;
         if (current_node->score < 0)
@@ -197,19 +202,18 @@ int is_spawn_allowed(server_rec * main_server, fcgid_command * command)
 
         /* Score is higher than up limit? */
         if (current_node->score >= sconf->spawnscore_uplimit) {
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0, main_server,
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, main_server,
                          "mod_fcgid: %s spawn score %d >= %d, skip the spawn request",
                          command->cgipath, current_node->score,
                          sconf->spawnscore_uplimit);
             return 0;
         }
 
-        /* 
-           Process count of this class higher than up limit?
+        /*
+         * Process count of this class higher than up limit?
          */
-        /* I need max class proccess count */
         if (current_node->process_counter >= current_node->max_class_process_count) {
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0, main_server,
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, main_server,
                          "mod_fcgid: too many %s processes (current:%d, max:%d), skip the spawn request",
                          command->cgipath, current_node->process_counter,
                          current_node->max_class_process_count);
